#include <MIDI.h>

// including this define cancels passthrough of realtime events.
// comment this out if using the device in series with an external
// synth
// #define NOLOOPBACK

#define ledPin 13
#define CHANNEL 1 // MIDI channel to use for sequencer - others are passed through
#define STAT1 7
#define STAT2 6

#define D2 2
#define D3 3
#define D4 4


   short int countStep; // the step number (number of frames in step is variable)
   short int countFrame;// the frame number (frame=external clock or 24 frames per quarter note)
   unsigned int newFrame; // has the sequencer hit a new frame - if so how many to deal with
   unsigned int newStep; // has the sequence hit a new step - if so how many to deal with
	// don't expect newFrame or newStep to exceed 1 but helps to get the sequence to catch up
	// if things start to run behind 
   
   int seqLength; // the number of steps the sequence is looped for - can be less than maxSeqLength
   int currentSeq; // the currently selected sequence row from the big matrix below
   int framesPerStep; 	// how many frames to count before the sequencer moves to another step.
			// when under external clock, the 'tempo' pot sets frames per step in 3's
			// steps are then:
			// 3fps=32nd notes, 6fps=16th notes (default), 9fps=dotted 16th, 12fps=8th notes
			// 15fps=something, 18fps=dotted 8th.

   const int maxSeqLength=16;	// the total possible number of steps (columns in the big matrix below)
   const int numberSequences=8;	// the number of rows in the big matrix below
   
   
	// this is the sequence data - 0 is no note, 1-4 are the 'activeNotes'
	// the first (index 0) sequence stores the 'roland mode' note data - these are actual MIDI notes 0 to 127
   int rolandVelo[maxSeqLength];
   int noteSeq[numberSequences][maxSeqLength] = {	{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
   																{1,0,2,0,3,0,4,0,5,0,6,0,7,0,8,0},
   																{1,0,3,0,5,0,7,0,2,0,4,0,6,0,8,0},
   																{1,2,0,8,7,0,3,4,0,6,5,0,6,7,8,0},
   																{1,3,5,2,0,0,0,0,8,6,4,7,0,0,0,0},
   																{1,0,0,3,0,0,2,0,0,5,0,0,4,0,0,6},
   																{1,3,5,0,8,6,4,0,3,5,7,0,6,4,2,0},
   																{0,0,0,0,0,0,0,0,1,0,0,0,1,4,2,3}	};

   // data about notes being held on keyboard for arp mode
   int numNotesHeld; // how many notes are being held on the keyboard
   int numActiveNotes; // number of notes currently stored in cache
   const int holdNotes=8; // maximum number of notes that can be held at once
   short int activeNotes[holdNotes] = {-1,-1,-1,-1,-1,-1,-1,-1}; // contains the MIDI notes the sequencer plays
   short int activeVelo[holdNotes] = {-1,-1,-1,-1,-1,-1,-1,-1}; // velocity received for notes

   // booleans and numbers that govern the different states the sequencer may be in
   bool pentatonic;
   short unsigned int pentatonicNote[128];

   bool playing; // is the sequencer running
   bool internalClock; // is the sequencer using the internal clock (otherwise external sync)
   bool muted;	// temporarily suspend playing any notes from the sequencer
   bool bypass;	// disable all functions and pass all note data through
   int trigMode;	// governs how the sequencer responds to notes
			// 0=sequencer runs free
			// 1=sequencer is 'muted' unless some notes are held (note gate)
			// 2=as 1, but sequencer is also reset when new notes are pressed (note retrigger)

   int rolandLowNote;	// for 'roland' mode, the sequence is transposed based on the lowest note recorded
   int rolandTransposeNote;	// the last note number received for transposing the sequence
   bool recording;	// overdubbing new notes into the sequence?

   // 
   int noteMaxLength; // this number gets put in the cronLength buffer when the sequencer generates a note
   int noteDecrement; 	// this number is subtracted from the cronLength buffer for each frame a note is active
			// and can be changed with the pot
   short const int queueLength = 16; // the number of notes that can be played at once
   int cronNote[queueLength]; // stores the notes playing
   int cronLength[queueLength]; // stores the amount of time a note has left to play
   
   long unsigned int nextOCR1A; // the hardware interrupt clock cycle for internal clocking
   const int timeoutExceeds=256; // if clock signal is lost, switch back to internal  
   int clockTimeout; 

   // control stuff
   unsigned int newA0; // state tracking for knobs
   unsigned int oldA0;
   unsigned int newA1;
   unsigned int oldA1;
	
   bool newD2; // state tracking for buttons
   bool oldD2;
   bool newD3;
   bool oldD3;
   bool newD4;
   bool oldD4;
	
   bool stateD2; // pressed = true
   bool stateD3;
   bool stateD4;


   void setup() {
      // set pin states
      pinMode(ledPin, OUTPUT);
      pinMode(STAT1, OUTPUT);
      pinMode(STAT2, OUTPUT);
      pinMode(D2, INPUT);
      pinMode(D3, INPUT);
      pinMode(D4, INPUT);
     	
      digitalWrite(D2,HIGH);
      digitalWrite(D3,HIGH);
      digitalWrite(D4,HIGH);
   
      digitalWrite(STAT1,HIGH);
      digitalWrite(STAT2,HIGH);
   
      // initialize timer1 
      noInterrupts();           // disable all interrupts
      TCCR1A = 0;// set entire TCCR1A register to 0
      TCCR1B = 0;// same for TCCR1B
      TCNT1  = 0;//initialize counter value to 0
      // set compare match register
      OCR1A = 60000;
      // turn on CTC mode
      TCCR1B |= (1 << WGM12);
      // Set CS11 bits for 8 prescaler
      TCCR1B |= (1 << CS11);  
      // enable timer compare interrupt
      TIMSK1 |= (1 << OCIE1A);
      interrupts();             // enable all interrupts
   
      // define MIDI library callbacks
      MIDI.begin(CHANNEL);
      MIDI.setThruFilterMode(midi::DifferentChannel);
   
      MIDI.setHandleClock(HandleClock);
      MIDI.setHandleStart(HandleStart);
      MIDI.setHandleStop(HandleStop);
      
      MIDI.setHandleControlChange(HandleControlChange);
      MIDI.setHandleNoteOff(HandleNoteOff);
      MIDI.setHandleNoteOn(HandleNoteOn);
      MIDI.setHandlePitchBend(HandlePitchBend);
      MIDI.setHandleProgramChange(HandleProgramChange);
      
      MIDI.setHandleAfterTouchPoly(HandleAftertouchPoly);
   
      // init variables
      pentatonic=false;
      playing=false;
      internalClock=true;
      bypass=false;
      countStep=0;
      countFrame=0;
      seqLength=16;
      currentSeq=0;
      noteMaxLength=1024;
      noteDecrement=300; // set a short note initially
      framesPerStep=6;	// when under external clock, the 'tempo' pot sets frames per step in 3's
   		// steps are then:
   		// 3fps=32nd notes, 6fps=16th notes, 9fps=dotted 16th, 12fps=8th notes
   		// 15fps=something, 18fps=dotted 8th.
      newFrame=0;
      newStep=0;
      numNotesHeld=0;
      numActiveNotes=0;
      muted=0;
      trigMode=0;
      
      for (int i=0;i<queueLength;i++){
         cronNote[i]=-1;
         cronLength[i]=-1;
      }
   
      for (int i=0;i<maxSeqLength;i++){
         rolandVelo[i]=0;
      }
      
      blink(3);
   }

   // the interrupt routine acts as an internal clock, or a check to make sure the external clock
   // is still running
   ISR(TIMER1_COMPA_vect)
   {
      if(internalClock){
      
         countFrame=(countFrame+1)%framesPerStep;
         newFrame++;
         if(!countFrame&&playing){
            countStep=(countStep+1)%seqLength;
            newStep++;
         }
      }
      else{
         clockTimeout++;
         if (clockTimeout>timeoutExceeds){
            HandleStop();
         }
      }
   }

// these are here just to provide thru function for other MIDI data on CHANNEL
// data on omni and other channels is sent by setting thru to DifferentChannel
   void HandleControlChange(byte channel, byte number, byte value){
      MIDI.sendControlChange(number,value,channel);
   }
   
   void HandlePitchBend(byte channel, int bend){
      MIDI.sendPitchBend(bend, channel);
   }
   
   void HandleProgramChange(byte channel, byte number){
      MIDI.sendProgramChange(number, channel);
   }

   void HandleAftertouchPoly(byte channel, byte note, byte pressure){
      MIDI.sendPolyPressure(note, pressure, channel);
   }



// start message - switch over from internal clock and re-start the sequence
   void HandleStart () {
   
      if(!bypass){
         playing=true;
         internalClock=false;
         countStep=-1;
         countFrame=-1;
         newFrame=0;
         newStep=0;
      /*
         for (int i=0;i<queueLength;i++){
            cronNote[i]=-1;
            cronLength[i]=-1;
         }
      */
      
      }
   }


   void HandleClock () {
      if(!bypass){
         playing=true;
         internalClock=false;
         clockTimeout=0;
         countFrame=(countFrame+1)%framesPerStep;
         newFrame++;
         if(!countFrame){
            countStep=(countStep+1)%seqLength;
            newStep++;
         }
      
      }
   }

// de-cron all notes and send noteOffs, stop the sequence and switch over to internal clock
   void HandleStop() {
   
      if(!bypass){
         for (int i=0;i<queueLength;i++){
            if ( (cronLength[i]>-1)||(cronNote[i]>-1) ){
               MIDI.sendNoteOff(cronNote[i],0,CHANNEL);
               cronLength[i]=-1;
               cronNote[i]=-1;
            }
         
         }
      
         digitalWrite(STAT2,HIGH);
         playing=false;
         internalClock=true;
         numNotesHeld=0;
      }
   }

// handle a note event on CHANNEL
   void HandleNoteOn(byte channel, byte note, byte velocity){
      if (pentatonic){
         note=pentatonicNote[note];
      }
      if( (channel==CHANNEL)&&(!bypass) ){
         if(!playing&&internalClock&&recording){
            countStep=( (countStep+1)%maxSeqLength );
            if(!((countStep+1)%4)){
               digitalWrite(STAT2,LOW);
            }
            else{
               digitalWrite(STAT2,HIGH);
            }
            noteSeq[0][countStep]=note;
            rolandVelo[countStep]=velocity;
         
                   
            bool queued=0; // has current note been queued?
            int shortest=noteMaxLength;
            int shortIdx=0;
            for (int i=0;(i<queueLength)&&(!queued);i++){ // check avail queue
               if ( (cronLength[i]<0)||(cronNote[i]==note) ){ // free queue slot
                  MIDI.sendNoteOn(note,velocity,CHANNEL);
                  if (cronNote[i]>=0){ // contains a note that hasn't been collected yet
                     MIDI.sendNoteOff(cronNote[i],0,CHANNEL);
                  }
                  cronNote[i]=note;
                  cronLength[i]=noteMaxLength;
                  queued=true;
               }
               else {
                  if (cronLength[i]<shortest) {
                     shortest=cronLength[i];
                     shortIdx=i;
                  }
               }
            }
            if (!queued){ // no free queue slots, steal the next expiry
               MIDI.sendNoteOff(cronNote[shortIdx],0,CHANNEL);
               MIDI.sendNoteOn(note,velocity,CHANNEL);
               cronNote[shortIdx]=note;
               cronLength[shortIdx]=noteMaxLength;
               queued=true;
            }
         
         
         }
         if (!recording){   
            if(numNotesHeld<=0){ // clear the note cache
               for (int i=0;i<holdNotes;i++){
                  activeNotes[i]=-1;
                  activeVelo[i]=-1;
               }
            }
         
         // push the note cache
            for (int i=0;i<(holdNotes-1);i++){
               activeNotes[i]=activeNotes[i+1];
               activeVelo[i]=activeVelo[i+1];
            } 
         
         
            activeNotes[holdNotes-1]=note;
            activeVelo[holdNotes-1]=velocity;
         
         // count the number of total notes in cache
            numActiveNotes=0;
            for (int i=0;i<holdNotes;i++){
               if (activeNotes[i]>=0){
                  numActiveNotes++;
               }
            
            }
         
            if (trigMode>0){
               muted=false;
               if ( (trigMode==2)&&(numNotesHeld==1) ){
                  countStep=-1;
                  countFrame=-1;
               }
            }
         }
      
         if (recording&&playing){
            if ((countFrame+2)>(framesPerStep/2)){
               noteSeq[0][(countStep+1)%maxSeqLength]=note;
               rolandVelo[(countStep+1)%maxSeqLength]=velocity;
            }
            else {
               noteSeq[0][countStep]=note;
               rolandVelo[countStep]=velocity;
               bool queued=0; // has current note been queued?
               int shortest=noteMaxLength;
               int shortIdx=0;
               for (int i=0;(i<queueLength)&&(!queued);i++){ // check avail queue
                  if (cronLength[i]<0){ // free queue slot
                     MIDI.sendNoteOn(note,velocity,CHANNEL);
                     if (cronNote[i]>=0){ // contains a note that hasn't been collected yet
                        MIDI.sendNoteOff(cronNote[i],0,CHANNEL);
                     }
                     cronNote[i]=note;
                     cronLength[i]=noteMaxLength;
                     queued=true;
                  }
                  else {
                     if (cronLength[i]<shortest) {
                        shortest=cronLength[i];
                        shortIdx=i;
                     }
                  }
               }
               if (!queued){ // no free queue slots, steal the next expiry
                  MIDI.sendNoteOff(cronNote[shortIdx],0,CHANNEL);
                  MIDI.sendNoteOn(note,velocity,CHANNEL);
                  cronNote[shortIdx]=note;
                  cronLength[shortIdx]=noteMaxLength;
                  queued=true;
               }
            
            }
         }
         rolandTransposeNote=note;
         numNotesHeld++;
      }
      else {
         MIDI.sendNoteOn(note, velocity, channel);
      }
      if(stateD2&&stateD3&&stateD4){
         pentatonic=true;
         int step=0;
         int closeNote=note;
         for (int i=note;i<128;i++){
            if (step==0){
               if (i<closeNote+2){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=1;
                  closeNote=closeNote+2;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==1){
               if (i<closeNote+2){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=2;
                  closeNote=closeNote+2;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==2){
               if (i<closeNote+3){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=3;
                  closeNote=closeNote+3;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==3){
               if (i<closeNote+2){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=4;
                  closeNote=closeNote+2;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==4){
               if (i<closeNote+3){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=0;
                  closeNote=closeNote+3;
                  pentatonicNote[i]=closeNote;
               }
            }
         }
         step=0;
         closeNote=note-3;
         for (int i=note-1;i>=0;i--){
         
         
            if (step==0){
               if (i>closeNote-3){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=1;
                  closeNote=closeNote-3;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==1){
               if (i>closeNote-2){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=2;
                  closeNote=closeNote-2;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==2){
               if (i>closeNote-3){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=3;
                  closeNote=closeNote-3;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==3){
               if (i>closeNote-2){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=4;
                  closeNote=closeNote-2;
                  pentatonicNote[i]=closeNote;
               }
            }
            else if (step==4){
               if (i>closeNote-2){
                  pentatonicNote[i]=closeNote;
               }
               else {
                  step=0;
                  closeNote=closeNote-2;
                  pentatonicNote[i]=closeNote;
               }
            }
            
            if (pentatonicNote[i]<0){
               pentatonicNote[i]=0;
            }
         
         }
      
      }
   
   }
   
   void HandleNoteOff(byte channel, byte note, byte velocity){
      if(pentatonic){
         note=pentatonicNote[note];
      }
      if( (channel==CHANNEL)&&(!bypass) ){
         numNotesHeld--;
         if(numNotesHeld<=0){
            numNotesHeld=0;
            if (trigMode>0){
               muted=true;
            }
         }
      
      }  
      
      else {
         MIDI.sendNoteOff(note, velocity, channel);
      }
   }

   void loop() {
      MIDI.read();
      if(numNotesHeld>0){
         digitalWrite(ledPin, HIGH);
      }
      else{
         digitalWrite(ledPin, LOW);
      }  
   	
      while(newFrame>0){
         HandleFrame(countFrame);
         newFrame--;
      }
      while(newStep>0){
         HandleStep(countStep);
         newStep--;
      }
   }

   void HandleStep(int countStep){
      if (!bypass){
      // turn STAT2 on every 4th step
         if (!(countStep%4)&&(playing)){
            digitalWrite(STAT2,LOW);
            if(!countStep&&recording){
               digitalWrite(STAT1,HIGH);
            }
         }
         if ( (recording)&&(currentSeq==0)&&(stateD4) ){
            noteSeq[0][countStep]=-1;
            rolandVelo[countStep]=0;
         }
      
         if (currentSeq>0){ // normal arpeggiating sequencer
            if ( (noteSeq[currentSeq][countStep]>0)&&(numActiveNotes>0)&&(!muted)&&(playing) ){ // step has a note - enqueue and play
               int note=activeNotes[ (noteSeq[currentSeq][countStep]-1)%numActiveNotes +holdNotes-numActiveNotes];
               int velo=activeVelo[ (noteSeq[currentSeq][countStep]-1)%numActiveNotes +holdNotes-numActiveNotes];
               bool queued=0; // has current note been queued?
               int shortest=noteMaxLength;
               int shortIdx=0;
               for (int i=0;(i<queueLength)&&(!queued);i++){ // check avail queue
                  if (cronLength[i]<0){ // free queue slot
                     MIDI.sendNoteOn(note,velo,CHANNEL);
                     if (cronNote[i]>=0){ // contains a note that hasn't been collected yet
                        MIDI.sendNoteOff(cronNote[i],0,CHANNEL);
                     }
                     cronNote[i]=note;
                     cronLength[i]=noteMaxLength;
                     queued=true;
                  }
                  else {
                     if (cronLength[i]<shortest) {
                        shortest=cronLength[i];
                        shortIdx=i;
                     }
                  }
               }
               if (!queued){ // no free queue slots, steal the next expiry
                  MIDI.sendNoteOff(cronNote[shortIdx],0,CHANNEL);
                  MIDI.sendNoteOn(note,velo,CHANNEL);
                  cronNote[shortIdx]=note;
                  cronLength[shortIdx]=noteMaxLength;
                  queued=true;
               }
            }
         }
         else if (currentSeq==0){ // Roland mode
            if ( (noteSeq[currentSeq][countStep]>=0)&&(!muted)&&(playing) ){ // step has a note - enqueue and play
               int note;
               if(recording){
                  note=noteSeq[0][countStep];
               }
               else{
                  note=(noteSeq[0][countStep]+rolandTransposeNote-rolandLowNote);
               }
               int velo=rolandVelo[countStep];
            // if note gets transposed out of range, raise or lower an octave
               while (note>127){
                  note=note-12;
               }
               while (note<0){
                  note=note+12;
               }
            
               bool queued=0; // has current note been queued?
               int shortest=noteMaxLength;
               int shortIdx=0;
               for (int i=0;(i<queueLength)&&(!queued);i++){ // check avail queue
                  if (cronLength[i]<0){ // free queue slot
                     MIDI.sendNoteOn(note,velo,CHANNEL);
                     if (cronNote[i]>=0){ // contains a note that hasn't been collected yet
                        MIDI.sendNoteOff(cronNote[i],0,CHANNEL);
                     }
                     cronNote[i]=note;
                     cronLength[i]=noteMaxLength;
                     queued=true;
                  }
                  else {
                     if (cronLength[i]<shortest) {
                        shortest=cronLength[i];
                        shortIdx=i;
                     }
                  }
               }
               if (!queued){ // no free queue slots, steal the next expiry
                  MIDI.sendNoteOff(cronNote[shortIdx],0,CHANNEL);
                  MIDI.sendNoteOn(note,velo,CHANNEL);
                  cronNote[shortIdx]=note;
                  cronLength[shortIdx]=noteMaxLength;
                  queued=true;
               }
            }
         }
      }
   }   
    
   void HandleFrame(int countFrame){
      if (!bypass){
      // turn STAT2 off 2 steps after turning it on
         if ( (countFrame==2)&&(playing) ){
            digitalWrite(STAT2,HIGH);
            if(!countStep&&recording){
               digitalWrite(STAT1,LOW);
            }
         }
      
      // decrement crons
      
         for (int i=0;i<queueLength;i++){
            if (cronLength[i]>-1){
               cronLength[i]=cronLength[i]-noteDecrement;
            }
            if (cronLength[i]<-1){
               MIDI.sendNoteOff(cronNote[i],0,CHANNEL);
               cronLength[i]=-1;
               cronNote[i]=-1;
            }
         
         }
      
         if(internalClock){
            MIDI.sendRealTime(midi::Clock);
         }
      
      }
   	// get control data;
      updateControls();	
      
   }
	  
   void updateControls(){
    
    	
      newA0=analogRead(0);
      if(abs(newA0-oldA0)>8){
      
         if(!stateD2&&!stateD3&&!stateD4){ // no buttons held, A0 corresponds to pattern select
            currentSeq=floor( ( (1023-newA0)*numberSequences)/1024 );
            if (currentSeq>0){
               recording=false;
               digitalWrite(STAT1, HIGH);
            }
         }
       	
         if(!stateD2&&!stateD3&&stateD4&&!internalClock){ // D4 held, A0 corresponds to note divisions, or 
         
            framesPerStep=floor(( floor(  ((1023-newA0)*15)/1024  )+1 )/3)*3 + 3;
         }
         if(!stateD2&&!stateD3&&stateD4&&internalClock){ // D4 held, A0 corresponds to note divisions, or 
         
            OCR1A=((1023-newA0)*48)+16431;
         }
      	
         oldA0=newA0;
      }
   	  
   	
      newA1=analogRead(1);
      if(abs(newA1-oldA1)>8){
      
         if(!stateD2&&!stateD3&&!stateD4){ // no buttons held, A1 corresponds to note length
            noteDecrement=newA1+24;
         }
      
         if(!stateD2&&!stateD3&&stateD4){ // D4 held, A1 corresponds to pattern length
            seqLength=floor( ( (1023-newA1)*maxSeqLength)/1024 )+1;
         }
         
         if(!stateD2&&stateD3&&!stateD4){ // D3 held, A1 correponds to trigger mode
            trigMode=floor( ( (1023-newA1)*3)/1024);
            if (trigMode>0){
               muted=true;
            }
            else{
               muted=false;
            }
         }
         oldA1=newA1;
         
      }
   	  
   
   
   	// D2 resets the sequencer
      newD2=!digitalRead(D2);
      if (newD2&&!oldD2){
      
         stateD2=true;
         if(stateD3&&stateD4){
            pentatonic=false;
         }
         if(!bypass){
            if (stateD4&&!stateD3){ // D4+D2=temporary mute
               if (trigMode==0){
                  muted=true;
               }
               if (trigMode>0){
                  muted=false;
               }
            }  
            else if ( (!stateD4)&&(stateD3)&&(currentSeq==0) ){ // D3+D2=record in Roland Mode
               if (recording){
                  recording=false;
                  digitalWrite(STAT1,HIGH);
               // get the low note for transpose
                  rolandLowNote=128;
                  for (int i=0;i<maxSeqLength;i++){
                     if ( (noteSeq[0][i]<rolandLowNote)&&(noteSeq[0][i]>=0) ){
                        rolandLowNote=noteSeq[0][i];
                     }
                  }
                  rolandTransposeNote=rolandLowNote;
                  if (internalClock&&!playing){
                     countStep=-1;
                     countFrame=-1;
                     digitalWrite(STAT2,LOW);
                  }
               }
               else {
                  recording=true;
                  rolandTransposeNote=rolandLowNote;
                  digitalWrite(STAT1,LOW);
                  if (internalClock&&!playing){
                     countStep=-1;
                     countFrame=-1;
                     digitalWrite(STAT2,LOW);
                  }
               }
            }
            else{ // D2 is sequencer reset or start/stop under internal clock
               countStep=-1;
               countFrame=-1;
            
               if(internalClock){
                  playing=!playing;
                  framesPerStep=6;
                  if(playing){
                     MIDI.sendRealTime(midi::Start);
                  }
                  else {
                     MIDI.sendRealTime(midi::Start);
                  }
               }
            }
         }
      }
      else if (oldD2&&!newD2){
         stateD2=false;
         
         if (stateD4){ // D4+D2=temporary mute (unmute on D2 up)
            if (trigMode==0){
               muted=false;
            }
            if (trigMode>0){
               muted=true;
            }
         } 
      }
      
   	
      newD3=!digitalRead(D3);
      if (newD3&&!oldD3){
         stateD3=true;
      }
      else if (oldD3&&!newD3){
         stateD3=false;
      }
     
      
      newD4=!digitalRead(D4);
      if (newD4&&!oldD4){
         stateD4=true;
         if(stateD3){
         
            if(bypass){
               bypass=false;
               digitalWrite(STAT2,HIGH);
            }
            else{
               bypass=true;
               digitalWrite(STAT2,LOW);
            }
         }
         if(!playing&&internalClock&&recording){
            countStep++;
            if(!((countStep+1)%4)){
               digitalWrite(STAT2,LOW);
            }
            else{
               digitalWrite(STAT2,HIGH);
            }
            noteSeq[0][countStep]=-1;
            rolandVelo[countStep]=0;
         }
      }
      else if (oldD4&&!newD4){
         stateD4=false;
      }
   
      oldD2=newD2;
      oldD3=newD3;
      oldD4=newD4;
   	
    
   }
	  
   void blink(int numtimes){
      while (numtimes){
         delay(100);
         digitalWrite(STAT1,LOW);
         delay(100);
         digitalWrite(STAT1,HIGH);
         numtimes--;
      }
   }
      
      
      



